from serotiny.config.utils import add_config as add
from serotiny.config.utils import create_config as get
from serotiny.config.utils import merge_config as merge
from serotiny.config.utils import encrypt



from .continuous_bernoulli import CBLogLoss
from .weibull import WeibullLogLoss
from .cosine_loss import CosineLoss
from .adversarial_loss import AdversarialLoss
from .gaussian_nllloss import GaussianNLLLoss



from torch.nn.modules.loss import _Loss as Loss


class AdversarialLoss(Loss):
    def __init__(self, discriminator, loss, argmax=False, reduction="mean"):
        super(AdversarialLoss, self).__init__(None, None, reduction)
        self.discriminator = discriminator
        self.loss = loss
        self.argmax = argmax

    def forward(self, input, target):
        yhat = self.discriminator(input)
        if not self.argmax:
            loss = self.loss(yhat, target)
        else:
            loss = self.loss(yhat, target.argmax(1))

        # reduce across batch dimension
        if self.reduction == "none":
            return loss
        elif self.reduction == "sum":
            return loss.sum()
        elif self.reduction == "mean":
            return loss.mean()
        else:
            raise NotImplementedError(f"Unavailable reduction type: {self.reduction}")



import torch.nn.functional as F
from torch.nn.modules.loss import _Loss as Loss
import torch


class GaussianNLLLoss(Loss):
    def __init__(self, reduction="mean"):
        super(GaussianNLLLoss, self).__init__(None, None, reduction)
        self.reduction = reduction
        self.loss = torch.nn.GaussianNLLLoss(reduction=self.reduction)

    def forward(self, input, target):
        dim = target.size(1) // 2
        recon_means = target[:, :dim]
        recon_vars = F.softplus(target[:, dim:]).add(1e-3)
        target = torch.cat([recon_means, recon_vars], dim=1)
        loss = self.loss(recon_means, input, recon_vars)
        return loss



import gc
import inspect
import logging
from typing import Sequence, Union
from omegaconf import ListConfig, DictConfig, OmegaConf

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.utilities.parsing import get_init_args

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]
logger = logging.getLogger("lightning")
logger.propagate = False


def _is_primitive(value):
    if isinstance(value, (type(None), bool, str, int, float)):
        return True

    if isinstance(value, (tuple, list)):
        return all(_is_primitive(el) for el in value)

    if isinstance(value, dict):
        return all(_is_primitive(el) for el in value.values())

    return False


def _cast_omegaconf(value):
    if isinstance(value, (ListConfig, DictConfig)):
        return OmegaConf.to_container(value, resolve=True)
    return value


def _parse_init_args(frame):
    init_args = get_init_args(frame)
    if frame.f_back.f_code.co_name == "__init__":
        # if this was called from a subclass's init
        init_args.update(get_init_args(frame.f_back))

    init_args = {k: _cast_omegaconf(v) for k, v in init_args.items()}
    ignore = [arg for arg, v in init_args.items() if not _is_primitive(v)]
    if "optimizer" in ignore:
        ignore.remove("optimizer")

    for arg in ignore:
        del init_args[arg]
    return init_args


class BaseModel(pl.LightningModule):
    def __init__(self, **kwargs):
        super().__init__()

        frame = inspect.currentframe()
        init_args = _parse_init_args(frame)

        self.save_hyperparameters(init_args, logger=False)
        self.optimizer = init_args.pop("optimizer", torch.optim.Adam)
        self.lr_scheduler = init_args.pop("lr_scheduler", None)
        self.cache_outputs = init_args.get("cache_outputs", ("test",))
        self._cached_outputs = dict()

    def parse_batch(self, batch):
        raise NotImplementedError

    def forward(self, x, **kwargs):
        raise NotImplementedError

    def _step(self, stage, batch, batch_idx, logger):
        """Here you should implement the logic for a step in the
        training/validation/test process. The stage (training/validation/test)
        is given by the variable `stage`.

        Example:

        x = self.parse_batch(batch)

        if self.hparams.id_label is not None:
           if self.hparams.id_label in batch:
               ids = batch[self.hparams.id_label].detach().cpu()
               results.update({
                   self.hparams.id_label: ids,
                   "id": ids
               })

        return results
        """
        raise NotImplementedError

    def training_step(self, batch, batch_idx):
        return self._step("train", batch, batch_idx, logger=True)

    def validation_step(self, batch, batch_idx):
        return self._step("val", batch, batch_idx, logger=True)

    def test_step(self, batch, batch_idx):
        return self._step("test", batch, batch_idx, logger=False)

    def predict_step(self, batch, batch_idx):
        return self._step("predict", batch, batch_idx, logger=False)

    def _epoch_end(self, split, outputs):
        if split in self.cache_outputs:
            if split in self._cached_outputs:
                del self._cached_outputs[split]
                gc.collect()
            self._cached_outputs[split] = outputs

    def train_epoch_end(self, outputs):
        self._epoch_end("train", outputs)

    def validation_epoch_end(self, outputs):
        self._epoch_end("val", outputs)

    def test_epoch_end(self, outputs):
        self._epoch_end("test", outputs)

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters())
        if self.lr_scheduler is not None:
            scheduler = self.lr_scheduler(optimizer=optimizer)
            return {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": scheduler,
                    "monitor": "val_loss",
                    "frequency": 1,
                },
            }
        return optimizer



from typing import Optional, Sequence

import torch
import torch.nn as nn

# from pytorch3d.transforms.rotation_conversions import euler_angles_to_matrix

from serotiny.networks.layers import conv_block


def make_parallelepipeds(input_dims, batch_size=0):

    ranges = [torch.linspace(-1, 1, dim) for dim in input_dims]

    ppiped = torch.stack(torch.meshgrid(*ranges))

    if batch_size:
        repeats = (len(input_dims) + 1) * [1]
        return ppiped.unsqueeze(0).repeat(batch_size, *repeats)
    return ppiped


class ImplicitDecoder(nn.Module):
    def __init__(
        self,
        latent_dims,
        hidden_channels: Sequence[int],
        non_linearity: Optional[nn.Module] = None,
        final_non_linearity: Optional[nn.Module] = None,
        mode: str = "3d",
    ):
        super().__init__()
        self.mode = mode
        self._mode = 3 if mode == "3d" else 2
        self.latent_dims = latent_dims

        if non_linearity is None:
            non_linearity = nn.ReLU()

        if final_non_linearity is None:
            final_non_linearity = nn.Identity()

        self.final_non_linearity = final_non_linearity

        layers = []
        _in_channels = latent_dims
        for index, out_channels in enumerate(hidden_channels):
            layers.append(
                conv_block(
                    _in_channels + self._mode + (self.latent_dims if index != 0 else 0),
                    out_channels,
                    kernel_size=1,
                    up_conv=False,
                    non_linearity=non_linearity,
                    mode=mode,
                )
            )
            _in_channels = out_channels

        self.layers = nn.ModuleList(layers)

    def forward(self, z, input_dims=(10, 10, 10), angles=None, translations=None):
        ppipeds = make_parallelepipeds(input_dims, batch_size=z.shape[0])

        if angles is not None:
            pass
            #  rot_matrices = euler_angles_to_matrix(angles, convention="XYZ")
            #  ppipeds = torch.matmul(ppipeds, rot_matrices)

        if translations is not None:
            pass
            #  ppipeds = ppipeds + translations.view(
            #      x.shape[0], 1, *([1] * len(input_dims))
            #  )

        ppipeds = ppipeds.to(z.device)

        z = z.view(*z.shape, *([1] * len(input_dims)))
        z = z.expand(-1, -1, *input_dims)

        y = z
        # Batch,channel (lt dim),input dims
        for index, layer in enumerate(self.layers):
            to_cat = (ppipeds, y) if index == 0 else (ppipeds, z, y)
            res = layer(torch.cat(to_cat, axis=1))
            # if output and input dimensions match
            if res.shape[1] == y.shape[1]:
                # skip connection, excluding the scoordinates and the latent code
                y = res + y
            else:
                y = res

        return self.final_non_linearity(y)



from .image_vae import ImageVAE
from .tabular_vae import TabularVAE
from .implicit_decoder_vae import ImplicitDecoderVAE

# from .tabular_conditional_vae import TabularConditionalVAE
# from .tabular_condprior_vae import TabularConditionalPriorVAE



import re
from typing import Optional, Sequence, Union

import numpy as np
import pandas as pd


def split_dataframe(
    dataframe: pd.DataFrame,
    train_frac: float,
    val_frac: Optional[float] = None,
    return_splits: bool = True,
):
    """Given a pandas dataframe, perform a train-val-test split and either return three
    different dataframes, or append a column identifying the split each row belongs to.

    TODO: extend this to enable balanced / stratified splitting

    Parameters
    ----------
    dataframe: pd.DataFrame
        Input dataframe

    train_frac: float
        Fraction of data to use for training. Must be <= 1

    val_frac: Optional[float]
        Fraction of data to use for validation. By default,
        the data not used for training is split in half
        between validation and test

    return_splits: bool = True
        Whether to return the three splits separately, or to append
        a column to the existing dataframe and return the modified
        dataframe
    """

    # import here to optimize CLIs / Fire usage
    from sklearn.model_selection import train_test_split

    train_ix, val_test_ix = train_test_split(
        dataframe.index.tolist(), train_size=train_frac
    )
    if val_frac is not None:
        val_frac = val_frac / (1 - train_frac)
    else:
        # by default use same size for val and test
        val_frac = 0.5

    val_ix, test_ix = train_test_split(val_test_ix, train_size=val_frac)

    if return_splits:
        return dict(
            train=dataframe.loc[train_ix],
            valid=dataframe.loc[val_ix],
            test=dataframe.loc[test_ix],
        )

    dataframe.loc[train_ix, "split"] = "train"
    dataframe.loc[val_ix, "split"] = "valid"
    dataframe.loc[test_ix, "split"] = "test"

    return dataframe


def filter_rows(
    dataframe: pd.DataFrame, column: str, values: Sequence, exclude: bool = False
):
    """Filter a dataframe, keeping only the rows where a given column's value is
    contained in a list of values.

    Parameters
    ----------
    dataframe: pd.DataFrame
        Input dataframe

    column: str
        The column to be used for filtering

    values: Sequence
        List of values to filter for
    """

    if exclude:
        return dataframe.loc[~dataframe[column].isin(values)]
    else:
        return dataframe.loc[dataframe[column].isin(values)]


def _filter_columns(
    columns_to_filter: Sequence[str],
    regex: Optional[str] = None,
    startswith: Optional[str] = None,
    endswith: Optional[str] = None,
    contains: Optional[str] = None,
    excludes: Optional[str] = None,
) -> Sequence[str]:
    """Filter a list of columns, using a combination of different queries, or a `regex`
    pattern. If `regex` is supplied it takes precedence and the remaining arguments are
    ignored. Otherwise, the logical AND of the supplied filters is applied, i.e. the
    columns that respect all of the supplied conditions are returned.

    Parameters
    ----------
    columns_to_filter: Sequence[str]
        List of columns to filter

    regex: Optional[str] = None
        A string containing a regular expression to be matched

    startswith: Optional[str] = None
        A substring the matching columns must start with

    endswith: Optional[str] = None
        A substring the matching columns must end with

    contains: Optional[str] = None
        A substring the matching columns must contain

    excludes: Optional[str] = None
        A substring the matching columns must not contain
    """
    if regex is not None:
        return [col for col in columns_to_filter if re.match(regex, col)]

    keep = [True] * len(columns_to_filter)
    for i in range(len(columns_to_filter)):
        if startswith is not None:
            keep[i] &= str(columns_to_filter[i]).startswith(startswith)
        if endswith is not None:
            keep[i] &= str(columns_to_filter[i]).endswith(endswith)
        if contains is not None:
            keep[i] &= contains in str(columns_to_filter[i])
        if excludes is not None:
            keep[i] &= excludes not in str(columns_to_filter[i])

    return [col for col, keep_column in zip(columns_to_filter, keep) if keep_column]


def filter_columns(
    input: Union[pd.DataFrame, Sequence[str]],
    columns: Optional[Sequence[str]] = None,
    startswith: Optional[str] = None,
    endswith: Optional[str] = None,
    contains: Optional[str] = None,
    excludes: Optional[str] = None,
    regex: Optional[str] = None,
):
    """Select columns in a dataset, using different filtering options. See
    serotiny.data.dataframe.transforms.filter_columns for more details.

    Parameters
    ----------
    input: Union[pd.DataFrame, Sequence[str]]
        The input to operate on. It can either be a pandas DataFrame,
        in which case the result is a DataFrame with only the columns
        that match the filters; or it can be a list of strings, and
        in that case the result is a list containing only the strings
        that match the filters

    columns: Optional[Sequence[str]] = None
        Explicit list of columns to include. If it is supplied,
        the remaining filters are ignored

    startswith: Optional[str] = None
        A substring the matching columns must start with

    endswith: Optional[str] = None
        A substring the matching columns must end with

    contains: Optional[str] = None
        A substring the matching columns must contain

    excludes: Optional[str] = None
        A substring the matching columns must not contain

    regex: Optional[str] = None
        A string containing a regular expression to be matched
    """

    if isinstance(input, pd.DataFrame):
        if columns is None:
            columns = _filter_columns(
                input.columns.tolist(), regex, startswith, endswith, contains, excludes
            )
            return input[columns]
    return _filter_columns(
        input.columns.tolist(), regex, startswith, endswith, contains, excludes
    )


def sample_n_each(
    dataframe: pd.DataFrame,
    column: str,
    number: int = 1,
    force: bool = False,
    seed: int = 42,
):
    """Transform a dataframe to have equal number of rows per value of `column`.

    In case a given value of `column` has less than `number` corresponding rows:
    - if `force` is True the corresponding rows are sampled with replacement
    - if `force` is False all the rows are given for that value

    Parameters
    ----------
    dataframe: pd.DataFrame
        Input dataframe

    column: str
        The column to be used for selection

    number: int
        Number of rows to include per unique value of `column`

    force: bool = False
        Toggle upsampling of classes with number of samples smaller
        than `number`

    seed: int
        Random seed used for sampling
    """

    values = dataframe[column].unique()

    subsets = []
    for value in values:
        class_rows = dataframe[dataframe[column] == value]
        if force or (len(class_rows) >= number):
            subsets.append(
                class_rows.sample(
                    number,
                    random_state=seed,
                    # only sample with replacement if there
                    # aren't enough data points in this class
                    replace=(len(class_rows) < number),
                )
            )
        else:
            subsets.append(class_rows.sample(frac=1, random_state=seed))

    return pd.concat(subsets)


def append_one_hot(dataframe: pd.DataFrame, column: str):
    """Modifies its argument by appending the one hot encoding columns into the given
    dataframe. Calls function one_hot_encoding.

    Parameters
    -----------
    dataframe: pd.DataFrame
        Input dataframe

    column: str
        Column to convert into one hot encoding
    """

    # import here to optimize CLIs / Fire usage
    from sklearn.preprocessing import OneHotEncoder

    one_hot = OneHotEncoder(sparse=False).fit_transform(dataframe[[column]])

    for idx in range(one_hot.shape[1]):
        dataframe[f"{column}_one_hot_{idx}"] = one_hot[:, idx]

    return dataframe


def append_labels_to_integers(dataframe: pd.DataFrame, column: str):
    """Modifies its argument by appending the integer-encoded values of `column` into
    the given dataframe.

    Parameters
    -----------
    dataframe: pd.DataFrame
        Input dataframe

    column: str
        Column to convert into one hot encoding
    """

    # import here to optimize CLIs / Fire usage
    from sklearn.preprocessing import LabelEncoder

    dataframe[f"{column}_integer"] = LabelEncoder().fit_transform(dataframe[[column]])

    return dataframe


def append_class_weights(dataframe: pd.DataFrame, column: str):
    """Add class weights (based on `column`) to a dataframe.

    Parameters
    -----------
    dataframe: pd.DataFrame
        Input dataframe

    column: str
        Column to base the weights on
    """
    labels_unique, counts = np.unique(dataframe[column], return_counts=True)
    class_weights = [sum(counts) / c for c in counts]
    class_weights_dict = dict(zip(labels_unique, class_weights))
    weights = [class_weights_dict[e] for e in dataframe[column]]
    dataframe[f"{column}_class_weights"] = weights
    return dataframe


def make_random_df(columns: Sequence[str] = list("ABCD"), n_rows: int = 100):
    """Generate a random dataframe. Useful to test data wrangling pipelines.

    Parameters
    ----------
    columns: Sequence[str] = ["A","B","C","D"]
        List of columns to add to the random dataframe. If none are provided,
        a dataframe with columns ["A","B","C","D"] is created

    n_rows: int = 100
        Number of rows to create for the random dataframe
    """

    data = np.random.randn(n_rows, len(columns))
    return pd.DataFrame(data, columns=columns)



import collections
import json
import os

import yaml


def deep_merge(template, merge):
    """Recursive dict merge, combines values that are lists.

    This mutates template - the contents of merge are added to template
    (which is also returned).
    If you want to keep template you could call it like
    deep_merge(dict(template), merge)
    """
    for k, v in merge.items():
        if (
            k in template
            and isinstance(template[k], dict)
            and isinstance(merge[k], collections.Mapping)
        ):
            deep_merge(template[k], merge[k])
        # Removed option to extend the list if there are existing elements:
        # elif k in template and isinstance(template[k], list) and isinstance(v, list):
        #     template[k].extend(v)
        else:
            template[k] = merge[k]
    return template


def merge_config(
    template_path="",
    merge="",
    output_path="",
):
    template_extension = os.path.splitext(template_path)[-1]
    with open(template_path, "r") as template_file:
        if template_extension == ".yaml":
            template = yaml.load(template_file)
        elif template_extension == ".json":
            template = json.load(template_file)
        else:
            raise Exception(f"format {template_extension} not supported")

    output_extension = os.path.splitext(output_path)[-1]
    with open(output_path, "w") as output_file:
        updated = deep_merge(template, merge)

        if output_extension == ".json":
            json.dump(updated, output_file, indent=4)
        elif output_extension == ".yaml":
            yaml.dump(updated, output_file)
        else:
            raise Exception(f"format {output_extension} not supported")



import re
from pathlib import Path
from upath import UPath
from typing import Optional, Sequence, Union

import pandas as pd
import pyarrow.parquet
import scanpy as sc
import anndata


def read_h5ad(path, include_columns=None):
    """Read an annData object stored in a .h5ad file.

    Parameters
    ----------

    path: Union[Path, str]
        Path to the .h5ad file

    include_columns: Optional[Sequence[str]] = None
        List of column names and/or regex expressions, used to only include the
        desired columns in the resulting dataframe.

    Returns
    -------
    annData
    """
    dataframe = sc.read(path)

    if include_columns is not None:
        columns = []
        for filter_ in include_columns:
            columns += filter_columns(dataframe.obs.columns.tolist(), regex=filter_)

        dataframe.obs = dataframe.obs[columns]
    return dataframe


def read_parquet(path, include_columns=None):
    """Read a dataframe stored in a .parquet file, and optionally include only
    the columns given by `include_columns`

    Parameters
    ----------

    path: Union[Path, UPath, str]
        Path to the .parquet file
    include_columns: Optional[Sequence[str]] = None
        List of column names and/or regex expressions, used to only include the
        desired columns in the resulting dataframe.

    Returns
    -------

    dataframe: pd.DataFrame
    """
    if include_columns is not None:
        schema = pyarrow.parquet.read_schema(path, memory_map=True)

        columns = []
        for filter_ in include_columns:
            columns += filter_columns(schema.names, regex=filter_)
    else:
        columns = None

    return pd.read_parquet(path, columns=columns)


def read_csv(path, include_columns=None):
    """Read a dataframe stored in a .csv file, and optionally include only the
    columns given by `include_columns`

    Parameters
    ----------

    path: Union[Path, UPath, str]
        Path to the .csv file
    include_columns: Optional[Sequence[str]] = None
        List of column names and/or regex expressions, used to only include the
        desired columns in the resulting dataframe.

    Returns
    -------

    dataframe: pd.DataFrame
    """

    dataframe = pd.read_csv(path)

    if include_columns is not None:
        columns = []
        for filter_ in include_columns:
            columns += filter_columns(dataframe.columns.tolist(), regex=filter_)

        dataframe = dataframe[columns]

    return dataframe


def read_dataframe(
    dataframe: Union[Path, UPath, str, pd.DataFrame],
    required_columns: Optional[Sequence[str]] = None,
    include_columns: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    """Load a dataframe from a .csv or .parquet file, or assert a given
    pd.DataFrame contains the expected required columns.

    Parameters
    ----------

    dataframe: Union[Path, UPath, str, pd.DataFrame]
        Either the path to the dataframe to be loaded, or a pd.DataFrame. Supported
        file types are .csv and .parquet

    required_columns: Optional[Sequence[str]] = None
        List of columns that the dataframe must contain. If these aren't found,
        a ValueError is thrown

    include_columns: Optional[Sequence[str]] = None
        List of column names and/or regex expressions, used to only include the
        desired columns in the resulting dataframe. For .parquet this also results
        in slower read latency.

    Returns
    -------

    dataframe: pd.DataFrame
    """

    if required_columns is not None and include_columns is not None:
        include_columns += required_columns

    if isinstance(dataframe, str):
        dataframe = UPath(dataframe)

    if isinstance(dataframe, (UPath, Path)):
        try:
            dataframe = dataframe.expanduser().resolve(strict=True)
        except (NotImplementedError, FileNotFoundError):
            pass

        if not dataframe.is_file():
            raise FileNotFoundError("Manifest file not found at given path")

        if dataframe.suffix == ".csv":
            dataframe = read_csv(dataframe, include_columns)
        elif dataframe.suffix == ".parquet":
            dataframe = read_parquet(dataframe, include_columns)
        elif dataframe.suffix == ".h5ad":
            dataframe = read_h5ad(dataframe, include_columns)
        else:
            raise TypeError("File type of provided manifest is not .csv " "or .parquet")

    elif isinstance(dataframe, pd.DataFrame):
        if include_columns is not None:
            columns = []
            for filter_ in include_columns:
                columns += filter_columns(dataframe.columns, regex=filter_)

            dataframe = dataframe[columns]
    elif isinstance(dataframe, anndata.AnnData):
        if include_columns is not None:
            columns = []
            for filter_ in include_columns:
                columns += filter_columns(dataframe.obs.columns, regex=filter_)

            dataframe.obs = dataframe.obs[columns]
    else:
        raise TypeError(
            f"`dataframe` must be either a pd.DataFrame or a path to "
            f"a file to load one. You passed {type(dataframe)}"
        )

    if isinstance(dataframe, anndata.AnnData):
        # Make dataframe out of anndata object
        X = dataframe.X.toarray()
        X = pd.DataFrame(X, columns=[f"X_{i}" for i in range(X.shape[1])]).reset_index(
            drop=True
        )
        index = pd.DataFrame(dataframe.obs_names)
        dataframe = dataframe.obs.reset_index(drop=True)
        dataframe = pd.concat([X, dataframe], axis=1)
        dataframe = pd.concat([index, dataframe], axis=1)

    if required_columns is not None:
        missing_columns = set(required_columns) - set(dataframe.columns)
        if len(missing_columns) > 0:
            raise ValueError(
                f"Some or all of the required columns were not "
                f"found on the given dataframe:\n{missing_columns}"
            )

    return dataframe


def filter_columns(
    columns_to_filter: Sequence[str],
    regex: Optional[str] = None,
    startswith: Optional[str] = None,
    endswith: Optional[str] = None,
    contains: Optional[str] = None,
    excludes: Optional[str] = None,
) -> Sequence[str]:
    """Filter a list of columns, using a combination of different queries, or a
    `regex` pattern. If `regex` is supplied it takes precedence and the
    remaining arguments are ignored. Otherwise, the logical AND of the supplied
    filters is applied, i.e. the columns that respect all of the supplied
    conditions are returned.

    Parameters
    ----------
    columns_to_filter: Sequence[str]
        List of columns to filter

    regex: Optional[str] = None
        A string containing a regular expression to be matched

    startswith: Optional[str] = None
        A substring the matching columns must start with

    endswith: Optional[str] = None
        A substring the matching columns must end with

    contains: Optional[str] = None
        A substring the matching columns must contain

    excludes: Optional[str] = None
        A substring the matching columns must not contain
    """
    if regex is not None:
        return [col for col in columns_to_filter if re.match(regex, col)]

    keep = [True] * len(columns_to_filter)
    for i in range(len(columns_to_filter)):
        if startswith is not None:
            keep[i] &= str(columns_to_filter[i]).startswith(startswith)
        if endswith is not None:
            keep[i] &= str(columns_to_filter[i]).endswith(endswith)
        if contains is not None:
            keep[i] &= contains in str(columns_to_filter[i])
        if excludes is not None:
            keep[i] &= excludes not in str(columns_to_filter[i])

    return [col for col, keep_column in zip(columns_to_filter, keep) if keep_column]



class Loader:
    """Abstract class for manifest loaders. Subclasses of this class shall implement
    specific mechanisms to load data from pandas dataframes.

    Instances of this class (and its subclasses) are callable objects, used by the
    `DataframeDataset` class, which always provides them with a dataframe row from which
    data is extracted
    """

    def __init__(self):
        pass

    def __call__(self, row):
        raise NotImplementedError



import torch

from .abstract_loader import Loader


class LoadClass(Loader):
    """Loader class, used to retrieve class values from the dataframe,"""

    def __init__(self, num_classes: int, y_encoded_label: str, binary: bool = False):
        """
        Parameters
        ----------
        num_classes: int
            Number of possible class values

        y_encoded_label: str
            Name of column containing the class value

        binary: bool
            Flag to determine whether to return class values or
            one-hot vectors

        """

        super().__init__()
        self.num_classes = num_classes
        self.binary = binary
        self.y_encoded_label = y_encoded_label

    def __call__(self, row):
        if self.binary:
            return torch.tensor([row[str(i)] for i in range(self.num_classes)])

        return torch.tensor(row[self.y_encoded_label])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hydra_plugins',
 'hydra_plugins.serotiny_search_path',
 'hydra_plugins.serotiny_shell_plugin',
 'serotiny',
 'serotiny.cli',
 'serotiny.cli._utils',
 'serotiny.cli.config_cli',
 'serotiny.cli.dataframe_cli',
 'serotiny.config',
 'serotiny.config.defaults',
 'serotiny.datamodules',
 'serotiny.io',
 'serotiny.io.dataframe',
 'serotiny.io.dataframe.loaders',
 'serotiny.losses',
 'serotiny.ml_ops',
 'serotiny.models',
 'serotiny.models.utils',
 'serotiny.models.vae',
 'serotiny.models.vae.priors',
 'serotiny.networks',
 'serotiny.networks.basic_cnn',
 'serotiny.networks.layers',
 'serotiny.networks.mlp',
 'serotiny.networks.utils',
 'serotiny.networks.vae',
 'serotiny.transforms',
 'serotiny.transforms.dataframe',
 'serotiny.transforms.image']

package_data = \
{'': ['*'],
 'serotiny.config.defaults': ['data/*',
                              'mlflow/*',
                              'model/*',
                              'model/vae/*',
                              'trainer/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'aicsimageio>=4.6.4,<5.0.0',
 'bfio>=2.3.0,<3.0.0',
 'fire>=0.4.0,<0.5.0',
 'frozendict>=2.3.2,<3.0.0',
 'fsspec>=2022.3.0,<2023.0.0',
 'hydra-core>=1.2.0,<2.0.0',
 'joblib>=1.1.0,<2.0.0',
 'makefun>=1.13.1,<2.0.0',
 'mlflow>=1.23.1,<2.0.0',
 'nbformat>=5.2.0,<6.0.0',
 'numpy>=1.18,<2.0',
 'ome-zarr>=0.6.0,<0.7.0',
 'omegaconf>=2.2.2,<3.0.0',
 'packaging>=20.9,<21.0',
 'pandas>=1.1,<2.0',
 'pip>=22.1.2,<23.0.0',
 'pyarrow>=7.0.0,<8.0.0',
 'pycryptodome>=3.14.1,<4.0.0',
 'pytorch-lightning>=1.6.0,<2.0.0',
 's3fs>=2022.3.0,<2023.0.0',
 'scanpy>=1.9.1,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'torch>=1.11.0,<2.0.0',
 'universal-pathlib>=0.0.20,<0.0.21']

extras_require = \
{'docs': ['sphinx', 'furo', 'm2r2']}

entry_points = \
{'console_scripts': ['serotiny = serotiny.cli.cli:main',
                     'serotiny.predict = serotiny.cli.cli:main',
                     'serotiny.test = serotiny.cli.cli:main',
                     'serotiny.train = serotiny.cli.cli:main']}

setup_kwargs = {
    'name': 'serotiny',
    'version': '0.0.3.dev202210042337',
    'description': 'A framework of tools to structure, configure and drive deep learning projects',
    'long_description': "# serotiny\n\nWhile going about the work of building deep learning projects, several simultaneous problems seemed to emerge:\n\n* How do we reuse as much work from previous projects as possible, and focus on building the part of the project that makes it distinct?\n* How can we automate the generation of new models that are based on existing models, but vary in a crucial yet non-trivial way?\n* When generating a multiplicity of related models, how can we keep all of the results, predictions, and analyses straight?\n* How can the results from any number of trainings and predictions be compared and integrated in an insightful yet generally applicable way?\n\nSerotiny arose from the need to address these issues and convert the complexity of deep learning projects into something simple, reproducible, configurable, and automatable at scale.\n\nSerotiny is still a work-in-progress, but as we go along the solutions to these problems become more clear. Maybe you've run into similar situations? We'd love to hear from you.\n\n## Overview\n\n`serotiny` is a framework and set of tools to structure, configure and drive deep\nlearning projects, developed with the intention of streamlining the lifecycle of\ndeep learning projects at [Allen Institute for Cell Science](https://www.allencell.org/).\n\nIt achieves this goal by:\n\n- Standardizing the structure of DL projects\n- Relying on the modularity afforded by this standard structure to make DL projects highly\n  configurable, using [hydra](https://hydra.cc) as the framework for configuration\n- Making it easy to adopt best-practices and latest-developments in DL infrastructure\n  by tightly integrating with\n    - [Pytorch Lightning](https://pytorchlightning.ai) for neural net training/testing/prediction\n    - [MLFlow](https://mlflow.org) for experiment tracking and artifact management\n\nIn doing so, DL projects become reproducible, easy to collaborate on and can\nbenefit from general and powerful tooling.\n\n## Getting started\n\nFor more information, check our [documentation](https://allencell.github.io/serotiny),\nor jump straight into our [getting started](https://allencell.github.io/serotiny/getting_started.html)\npage, and learn how training a DL model can be as simple as:\n\n``` sh\n\n$ serotiny train data=my_dataset model=my_model\n\n```\n\n## Authors\n\n- Guilherme Pires @colobas\n- Ryan Spangler @prismofeverything\n- Ritvik Vasan @ritvikvasan\n- Caleb Chan @calebium\n- Theo Knijnenburg @tknijnen\n- Nick Gomez @gomeznick86\n\n## Citing\n\nIf you find serotiny useful, please cite this repository as:\n\n```\nSerotiny Authors (2022). Serotiny: a framework of tools to structure, configure and drive deep learning projects [Computer software]. GitHub. https://github.com/AllenCellModeling/serotiny\nFree software: BSD-3-Clause\n```\n",
    'author': 'Guilherme Pires',
    'author_email': 'guilherme.pires@alleninstitute.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://allencell.github.io/serotiny',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Sequence
import numpy as np

from serotiny.transforms.dataframe.transforms import _filter_columns as filter_columns

from .abstract_loader import Loader


class LoadColumn(Loader):
    """Loader class, used to retrieve fields directly from dataframe
    columns."""

    def __init__(self, column: str, dtype: str = "float", unsqueeze: bool = True):
        """
        Parameters
        ----------
        column: str
            The column to retrieve

        """
        super().__init__()

        self.column = [column] if unsqueeze else column
        self.dtype = np.dtype(dtype).type
        self.unsqueeze = unsqueeze

    def __call__(self, row):
        return self.dtype(row[self.column])


class LoadColumns(Loader):
    """Loader class, used to retrieve fields directly from multiple dataframe
    columns, concatenating them into an array.

    It leverages `filter_columns` to enable using simple queries to select the
    columns to use.
    """

    def __init__(
        self,
        columns: Optional[Sequence[str]] = None,
        startswith: Optional[str] = None,
        endswith: Optional[str] = None,
        contains: Optional[str] = None,
        excludes: Optional[str] = None,
        regex: Optional[str] = None,
        dtype: str = "float",
    ):
        """
        Parameters
        ----------
        columns: Sequence[str]
            Explicit list of columns to include. If it is supplied,
            the remaining filters are ignored

        regex: Optional[str] = None
            A string containing a regular expression to be matched

        startswith: Sequence[str] = None
            A substring the matching columns must start with

        endswith: Sequence[str] = None
            A substring the matching columns must end with

        contains: Sequence[str] = None
            A substring the matching columns must contain

        excludes: Sequence[str] = None
            A substring the matching columns must not contain

        dtype: str = "float"
            dtype of the resulting array

        """
        super().__init__()
        self.columns = columns
        self.startswith = startswith
        self.endswith = endswith
        self.contains = contains
        self.excludes = excludes
        self.regex = regex
        self.dtype = dtype

        if columns is None:
            assert (
                (startswith is not None)
                or (endswith is not None)
                or (contains is not None)
                or (excludes is not None)
                or (regex is not None)
            )
        else:
            self.columns = list(columns)

    def _filter_columns(self, columns_to_filter):
        if self.columns is None:
            self.columns = filter_columns(
                columns_to_filter,
                self.regex,
                self.startswith,
                self.endswith,
                self.contains,
                self.excludes,
            )

        return self.columns

    def __call__(self, row):
        filtered_cols = self._filter_columns(row.index)
        return row[filtered_cols].astype(self.dtype)
